#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include "routingTableGen.h"



#ifdef DEVICE_K2H 
/* CSL SRIO Functional Layer */
#include <ti/csl/tistdtypes.h>
#include <ti/csl/csl_types.h>
#include <ti/csl/cslr_device.h>
#include <ti/csl/csl_srio.h>
#include <ti/csl/csl_srioAux.h>
#include <ti/csl/csl_srioAuxPhyLayer.h>


#include <ti/csl/csl_serdes.h>
#include <ti/csl/csl_serdes_srio.h>

#endif

#ifndef ENABLE_LOOP_1_4
/** NO LOOPS **/
/* Data structure representing depth within nodes of a cartridge. This is used to find routing information  after
    the cartridge-cartridge route is found, and finding a way to reach the cartridge */
DepthToNodes_t DepthToNodes[4][4] = {
  	        /* N1  */     /* N2  */     /* N3  */   /* N4  */  
  /*N1 */	{{SAME,0} ,{RIGHT,0}, {RIGHT,0}, {RIGHT,2}},
  /*N2 */ 	{{LEFT,0},  {SAME,0},  {RIGHT,1}, {RIGHT,1}},
  /*N3 */	{{LEFT,0},  {LEFT,0},   {SAME,0},   {RIGHT,1}},
  /*N4 */	{{LEFT,2},{LEFT,1},   {LEFT,1},    {SAME,0}}
};

#else
/** WITH LOOPS **/
/* Data structure representing depth within nodes of a cartridge. This is used to find routing information  after
    the cartridge-cartridge route is found, and finding a way to reach the cartridge */
DepthToNodes_t DepthToNodes[4][4] = {
                        /* N1  */     /* N2  */     /* N3  */   /* N4  */
  /*N1 */               {{SAME,0} ,{RIGHT,0}, {RIGHT,0}, {LEFT,1}},
  /*N2 */               {{LEFT,0},  {SAME,0},  {RIGHT,1}, {RIGHT,1}},
  /*N3 */               {{LEFT,0},  {LEFT,0},   {SAME,0},   {RIGHT,1}},
  /*N4 */               {{RIGHT,1},{LEFT,1},   {LEFT,1},    {SAME,0}}
};
#endif

/************************* Find the Routing Path from one cartridge to the another, recursively *********************
This uses a recursive depth(distance) finding algorithm emanating from one cartridge and spreading out to the
neighbouring cartridges recursively until the destination is reached. The depth (cartridge_src, cartridge_dest,direction) is 
noted on the way.
**********************************************************************************************************/
int find_Path(Topology_t *cluster, Cartridge_t *parent_cartridge,Cartridge_t *my_cartridge, Cartridge_t *dest_cartridge , 
                      int catridge_ids_so_far[],int num_cartridges_so_far, DepthToCartridge_t DepthToCartridge[][MAX_NUM_CARTRIDGES], Rbool *repeat_ret) {

   int depth[4]={MAX_DEPTH,MAX_DEPTH,MAX_DEPTH,MAX_DEPTH}; /* For each LEFT/RIGHT/UP/DOWN */
   int myID,destID,dir,i;
   Rbool repeat_node;
   int smallest_depth_so_far,depth_ret; 
  char dir_string[4][15]={"LEFT","RIGHT","UP","DOWN"};

   Cartridge_t * adjacent_cartridge[4];

#ifdef DEBUG_PRINTF
   printf("\n FIND PATH: At %s to Dest %s: From Parent %s",my_cartridge->name,dest_cartridge->name,parent_cartridge->name);
   printf("\n FIND PATH: %d Cartridges so far...",num_cartridges_so_far);
      fflush(stdout);	
   for(i=0;i<num_cartridges_so_far;i++) {
      printf("  %d ",catridge_ids_so_far[i]);
   }
   fflush(stdout);
#endif

   if(!my_cartridge->enabled) {
#ifdef DEBUG_PRINTF
      printf("\n   FIND PATH: %s Disabled..so, exit..",my_cartridge->name);
#endif
      return BLOCKED; 
   }

   if(cartridge_present(my_cartridge->cartridgeID_XY,catridge_ids_so_far,num_cartridges_so_far)) {
#ifdef DEBUG_PRINTF
        printf("\n   FIND PATH: %s in a Loop..so, exit..",my_cartridge->name);
#endif
        *repeat_ret=TRUE;
        return BLOCKED;
   } 


   if(my_cartridge==dest_cartridge) {
#ifdef DEBUG_PRINTF
    printf("\n    FIND PATH: Dest reached %s...",dest_cartridge->name);
     fflush(stdout);
#endif
      return 1;
    }
  
    myID= my_cartridge->cartridgeID_XY;
    destID= dest_cartridge->cartridgeID_XY;
   smallest_depth_so_far=MAX_DEPTH;      
   for(dir=0;dir<4;dir++)  {
     /* Check if the depth entries are valid, if so, we could just use the existing one, no need to recalculate */
     if(!DepthToCartridge[myID][destID].entries_per_dir[dir].valid  && my_cartridge->connections[dir].enabled) {
        /* Calculate the depth through all adjacent cartridges and find the least depth one */
        adjacent_cartridge[dir]=  find_Cartridge_by_coordinates(cluster, &my_cartridge->connections[dir].remoteCartridge);
        /* Check if the adjacent cartridge to branch out, is the one we came from. if so, skip it */
         if(adjacent_cartridge[dir] != parent_cartridge) {
             /* Check if the adjacent cartridge is enabled, if not this path is blocked so, return MAX_DEPTH */
             if(adjacent_cartridge[dir]->enabled) {
                   catridge_ids_so_far[num_cartridges_so_far]=my_cartridge->cartridgeID_XY;
                   /* Recursive call */
                     repeat_node=FALSE; 
                   depth[dir]=find_Path(cluster,my_cartridge,adjacent_cartridge[dir],dest_cartridge,catridge_ids_so_far,num_cartridges_so_far+1,DepthToCartridge,&repeat_node);
                   if(!repeat_node) {
                      DepthToCartridge[myID][destID].entries_per_dir[dir].depth=depth[dir]; 
                      DepthToCartridge[myID][destID].entries_per_dir[dir].valid=TRUE; 
#ifdef DEBUG_PRINTF
                       printf("\n FIND PATH: calculated DEPTH(%s, %s), [%s]   = %d",my_cartridge->name,dest_cartridge->name, dir_string[dir],depth[dir]);
                      fflush(stdout);
#endif
                  /* if the depth found just now is lower than the others, record the minimum */
                      if(DepthToCartridge[myID][destID].entries_per_dir[dir].depth <  DepthToCartridge[myID][destID].entries_per_dir[DepthToCartridge[myID]              [destID].min_depth_dir].depth) {
                       /* New Minimum found */
                         DepthToCartridge[myID][destID].min_depth_dir=dir;
#ifdef DEBUG_PRINTF
                         printf("\n FIND PATH: Newly Minimum Found!!!" );
#endif
                      }  

                      if(depth[dir]<smallest_depth_so_far) {
                          smallest_depth_so_far=depth[dir];
                      }
                   }  
               } else {
                    depth[dir]=MAX_DEPTH;
              } /* if adjacent->enabled */
       }/* if adjacent!=parent_cartridge */
     } else {
               /* Use the exisiting depth to destination , precalculated earlier */
#ifdef DEBUG_PRINTF
             printf("\n FIND PATH: Depth Already exists =  %d @ %s, so just use it...",DepthToCartridge[myID][destID].entries_per_dir[dir].depth, dir_string[dir]); 
#endif
             depth[dir]=DepthToCartridge[myID][destID].entries_per_dir[dir].depth;
             if(depth[dir]<smallest_depth_so_far) {
                    smallest_depth_so_far=depth[dir];
               }
      }
   }  /* For dir =0-3 */       
   
        if(smallest_depth_so_far==MAX_DEPTH) {
               depth_ret=MAX_DEPTH; /* BLOCKED */
        } else {
               depth_ret=smallest_depth_so_far+1;
        } 

       return(depth_ret);
}   

/************************* Calculate the depth from one cartridge to all others in all directions  *********************/
void generate_depth_to_cartridges(Topology_t *cluster,DepthToCartridge_t DepthToCartridges[][MAX_NUM_CARTRIDGES]) {

    int catridge_ids_so_far[MAX_NUM_CARTRIDGES],num_cartridges_so_far,dir,src_xy_id,dest_xy_id,smallest_depth_so_far,smallest_depth_index;
    Cartridge_t *my_cartridge,*adjacent_cartridge,*remote_cartridge,*dest_cartridge;
    char dir_string[4][15]={"LEFT","RIGHT","UP","DOWN"};
    Rbool repeat_node;
     int i,j,depth[4]={MAX_DEPTH, MAX_DEPTH, MAX_DEPTH, MAX_DEPTH}; /* For each LEFT/RIGHT/UP/DOWN */

    for(i=0;i<cluster->num_cartridges;i++) {
           my_cartridge=&cluster->cartridges[i];
            if(my_cartridge->enabled) {
                 catridge_ids_so_far[0]=my_cartridge->cartridgeID_XY;
                 /* Find the depth to every other cartridge,except this one in all four directions */
#ifdef DEBUG_PRINTF
                 printf("\n Depth Generation: %s  START",my_cartridge->name);
                 fflush(stdout);
#endif
                 for(j=0;j<cluster->num_cartridges;j++) {
  
                     dest_cartridge=&cluster->cartridges[j];
                     dest_xy_id=srio_cartridge_id_xy(dest_cartridge->coordinates.x,dest_cartridge->coordinates.y);
               
                      if(dest_cartridge->enabled) { 
#ifdef DEBUG_PRINTF
                       printf("\n Depth Generation: Depth from %s  to %s (# %d) ",my_cartridge->name, dest_cartridge->name,dest_xy_id);
                       fflush(stdout);
                   #endif

                         if(i!=j) {

                          for(dir=0;dir<4;dir++) {
                               if(my_cartridge->connections[dir].enabled) {
                                  adjacent_cartridge=  find_Cartridge_by_coordinates(cluster, &my_cartridge->connections[dir].remoteCartridge);
                                  repeat_node=FALSE; 
                                 depth[dir]=find_Path(cluster,my_cartridge,adjacent_cartridge,dest_cartridge,catridge_ids_so_far,1,DepthToCartridges,&repeat_node);
                                } else {
                                 depth[dir]=MAX_DEPTH;  
                               } 
                          }
                      }  else {
                          for(dir=0;dir<4;dir++) {
                               depth[dir]=0;
                          }
                     }

                      src_xy_id=srio_cartridge_id_xy(my_cartridge->coordinates.x,my_cartridge->coordinates.y);

#ifdef DEBUG_PRINTF
                        printf("\n Gen Depth: SRC(%d,%d, # %d ) to DEST (%d,%d, # %d)..",my_cartridge->coordinates.x,my_cartridge->coordinates.y,src_xy_id,
dest_cartridge->coordinates.x,dest_cartridge->coordinates.y,dest_xy_id); 
                        fflush(stdout);
#endif
                     
                     smallest_depth_so_far=MAX_DEPTH;  
                     smallest_depth_index=0;
                     for(dir=0;dir<4;dir++) {
#ifdef DEBUG_PRINTF
                        printf("\n Gen Depth: Depth[%d][%d]@%s=%d...",src_xy_id,dest_xy_id,dir_string[dir],depth[dir]); 
                        fflush(stdout);
#endif
                         DepthToCartridges[src_xy_id - XY_ID_BASE][dest_xy_id - XY_ID_BASE].entries_per_dir[dir].depth=depth[dir];
                         DepthToCartridges[src_xy_id - XY_ID_BASE][dest_xy_id- XY_ID_BASE].entries_per_dir[dir].valid=TRUE;
                         if(depth[dir]<=smallest_depth_so_far) {
                             /* Update the smallest depth */
                             smallest_depth_so_far=depth[dir];
                             smallest_depth_index=dir;
                         }
                      }  
#ifdef DEBUG_PRINTF
                       printf("\n Gen Depth: MIN Depth [%d][%d]@%s=%d...",src_xy_id,dest_xy_id,dir_string[smallest_depth_index],depth[smallest_depth_index]); 
#endif
                       DepthToCartridges[src_xy_id- XY_ID_BASE][dest_xy_id- XY_ID_BASE].min_depth_dir=smallest_depth_index;
                     }/* if enabled */  
                }  /* j */
          } /* if enabled */                      
      } /* num_cartridges */                      
}

/************************* Print Depth To Cartridges *******************************************************/
void print_depth_to_cartridges  (Topology_t *cluster,DepthToCartridge_t depthToCartridges[][MAX_NUM_CARTRIDGES],
                                               int num_cartridges) {
   int i,j,dir;
   char dir_string[5][15]={"LEFT","RIGHT","UP","DOWN","SAME"}; 
   int min_depth_dir,my_x,my_y,my_z,dest_x,dest_y,dest_z, my_cartridge_xy_id,dest_cartridge_xy_id;
   Cartridge_t *my_cartridge, *dest_cartridge;
   DepthToCartridge_t *depthToCartridge;
   printf("\n Printing Depth Cx-Cy[][]...");

/* Go through the IDs in a progressive manner. This way we could add it to the routing table */
    for(my_y=0;my_y<cluster->num_rows; my_y++) {
       for(my_x=0;my_x<cluster->num_columns; my_x++) {
          
          my_cartridge=&(elem_2darray(cluster->cartridges,my_x,my_y,cluster->num_columns));
          my_cartridge_xy_id=srio_cartridge_id_xy(my_x,my_y);
  
          if(my_cartridge->enabled) {

              for(dest_y=0;dest_y<cluster->num_rows; dest_y++) {
                 for(dest_x=0;dest_x<cluster->num_columns; dest_x++) {

                    dest_cartridge=&(elem_2darray(cluster->cartridges,dest_x,dest_y,cluster->num_columns));
         
                    if(dest_cartridge->enabled) { 
                      dest_cartridge_xy_id=srio_cartridge_id_xy(dest_x,dest_y);

                     depthToCartridge=&depthToCartridges[my_cartridge_xy_id- XY_ID_BASE][dest_cartridge_xy_id- XY_ID_BASE];
                     printf("\n");
                     for(dir=0;dir<4;dir++) {
                         printf("Depth[%s #%d ][%s #%d] @ %s = ",my_cartridge->name,my_cartridge_xy_id, 
                                                                                                     dest_cartridge->name,dest_cartridge_xy_id,
                                                                                                      dir_string[dir]);
                       if(depthToCartridge->entries_per_dir[dir].depth!=MAX_DEPTH) {
                             printf("%d",depthToCartridge->entries_per_dir[dir].depth);
                       } else {
                             printf("...");
                       }

                          if(dir==depthToCartridge->min_depth_dir) {
                              printf("  <----- MINIMUM DEPTH \n");
                         } else {
                             printf("\n");
                          } 

                     }
 
                      } /*if dest enabled */
                 } /* dest_x */
              }/* dest_y */

          } /* if enabled */
          
       } /* my_x */
     }/* my_y */
  return;
}

/************************* Init Depth To Cartridges *******************************************************/
void init_depth_to_cartridges  (DepthToCartridge_t depthToCartridges[MAX_NUM_CARTRIDGES][MAX_NUM_CARTRIDGES],
                                               int num_cartridges) {
   int i,j,dir;
   for(i=0;i<num_cartridges;i++) {
       for(j=0;j<num_cartridges;j++) {
          depthToCartridges[i][j].min_depth_dir=0;
          for(dir=0;dir<4;dir++) {
              depthToCartridges[i][j].entries_per_dir[dir].valid=FALSE;
              depthToCartridges[i][j].entries_per_dir[dir].depth=MAX_DEPTH;
          }
      }   
   }

}
/************************* Initialize routing table ******************************************************/
void init_routing_table  (RoutingTable_t *routingTable) {
   int i;
    routingTable->num_entries=0;

  for(i=0;i<8;i++) {
      routingTable->entries[i].lo=0;
      routingTable->entries[i].hi=0;
      routingTable->entries[i].port=0;
  }


}

/************************************************ Enable Cartridges by Name ***********************/
void enable_cartridges_by_name(Topology_t *topology, char active_cartridges[][10], int num_active_cartridges) {
   int i,j,dir,num_cartridges;
   Cartridge_t *this_cartridge;

   for(i=0;i<topology->num_cartridges;i++) {
     this_cartridge=&topology->cartridges[i]; /* Stored in linear fashion */
     for(j=0;j<num_active_cartridges;j++) {
          if(!strcmp(active_cartridges[j],this_cartridge->name)) {
               printf("\n Cluster: Enabling %s",this_cartridge->name);
               this_cartridge->enabled=TRUE;
               break;
          } 
    }   
  }
}

/************************************************ Enable_cartridges_by_PhysID **********************************/
void enable_cartridges_by_PhysID(Topology_t *topology, int active_cartridges[], int num_active_cartridges) {
   int i,j,dir,num_cartridges;
   Cartridge_t *this_cartridge;

   for(i=0;i<topology->num_cartridges;i++) {
     this_cartridge=&topology->cartridges[i]; /* Stored in linear fashion */
     for(j=0;j<num_active_cartridges;j++) {
          if(active_cartridges[j]==this_cartridge->cartridgeID_Phys) {
#ifdef DEBUG_PRINTF
               printf("\n Cluster: Enabling %s",this_cartridge->name);
#endif
               this_cartridge->enabled=TRUE;
               break;
          } 
    }   
  }
}
/************************************************  Reset Cluster **********************************/
void reset_cluster(Topology_t *topology) {
   int i,j,k;
   Cartridge_t *this_cartridge;

   for(i=0;i<topology->num_cartridges;i++) {
     this_cartridge=&topology->cartridges[i]; /* Stored in linear fashion */
     if(this_cartridge->enabled) {
         this_cartridge->enabled=FALSE;
         for(j=0;j<4;j++) {
          this_cartridge->nodes[j].routingTable.num_entries=0; 
          for(k=0;k<8;k++) {
            this_cartridge->nodes[j].routingTable.entries[k].lo=0; 
            this_cartridge->nodes[j].routingTable.entries[k].hi=0; 
            this_cartridge->nodes[j].routingTable.entries[k].port=0; 
          }
        }
     }
  }
}



/************************************************ Enable_cartridges_by_PhysID **********************************/
int print_cluster(Topology_t *topology)
{
      int port_connections_no,node_connection_no,cartridge_connection_no,node_no,cartridge_no,port_no,dir;
     char *c_name,*node_name;
       Cartridge_t *this_cartridge; 
        NodeInfo_t *this_node;
       char direction[4][15]={"LEFT","RIGHT","UP","DOWN"};

      printf("\n ------CLUSTER DISPLAY---------\n");
      printf("\n Num Rows= %d, Num Columns = %d\n",topology->num_rows, topology->num_columns);

     printf("\nEnabled Cartridges of this cluster shown below\n");

    for(cartridge_no=0;cartridge_no<topology->num_cartridges; cartridge_no++) {
          /* Display Cartridge Info */
          c_name=topology->cartridges[cartridge_no].name;
           this_cartridge=&topology->cartridges[cartridge_no];
 
           if(this_cartridge->enabled) {
             printf("\n   %s:    ENABLED= %d",c_name, this_cartridge->enabled); 
             printf("\n   %s:    Name: %s    ",c_name,this_cartridge->name);
             printf("\n   %s:    PhysID =%d  ",c_name,this_cartridge->cartridgeID_Phys);
             printf("\n   %s:    XY =(%d,%d)     ",c_name,this_cartridge->coordinates.x,this_cartridge->coordinates.y);
             printf("\n   %s:    cartridgeID_XY =%d  ",c_name,this_cartridge->cartridgeID_XY);
             printf("\n   %s:    Number of Nodes=%d",c_name,this_cartridge->num_nodes);

          /* Display Nodes */
         for(node_no=0;node_no< this_cartridge->num_nodes; node_no++) {
             this_node=&this_cartridge->nodes[node_no];
             node_name=this_node->name; 
             printf("\n\n   %s: PhysID= %d    XYZ =(%d,%d,%d)   , srioID # %d",this_node->name,this_node->nodeID_Phys,this_node->coordinates.x,this_node->coordinates.y,this_node->coordinates.z, this_node->srioID);
             printf("\n   %s:    Connections Initialized = %d",node_name,this_node->connections_per_dir_initialized);

           /* 
            for(port_connections_no=0;port_connections_no<4;port_connections_no++) {  
                 printf("\n   %s:  Port[%d] --> %s ",node_name, port_connections_no,this_node->port_connections[port_connections_no]);
            }
           */
            for(port_connections_no=0;port_connections_no< this_node->num_connections; port_connections_no++) {
             if(this_node->connections[port_connections_no].enabled) {
                  NodeInfo_t *connected_node =find_Node_by_coordinates(topology, &(this_node->connections[port_connections_no].remoteNode));
                  printf("\n   %s:  [%d] --> %s on Port %d",this_node->name, port_connections_no, connected_node->name,this_node->connections[port_connections_no].port_no);

               } else {
                  printf("\n   %s:  [%d] --> NONE ",this_node->name, port_connections_no);
        
               }
            } 

             print_routing_table(this_node,&this_node->routingTable);

           /* Node connections */
               printf("\n Find node connections by direction ..");
          for(dir=0;dir< 4; dir++) {
             if(this_node->connections_per_dir[dir].enabled) {
                  NodeInfo_t *connected_node =find_Node_by_coordinates(topology, &(this_node->connections_per_dir[dir].remoteNode));
                   printf("\n   %s:  On %s  Port %d connected to %s",
                                                this_node->name, 
                                                direction[dir], 
                                                this_node->connections_per_dir[dir].port_no,
                                                connected_node->name);
               } else {
                   printf("\n   %s:  On %s  Port %d connected to NONE",
                                                this_node->name, 
                                                direction[dir],  
                                                this_node->connections_per_dir[dir].port_no);
        
               }
            } 
         } /* for node=0;node<4;*/

        /* Cartridge connections */
        printf("\n Cartridge Connections");
        for(dir=0;dir< 4; dir++) {
                Cartridge_t *connected_cartridge =find_Cartridge_by_coordinates(topology, &(this_cartridge->connections[dir].remoteCartridge));
                printf("\n %s connected on %s  to %s (%d,%d,%d)  , through port %d, via node %s  ",this_cartridge->name,
               direction[dir],
               connected_cartridge->name,
               this_cartridge->connections[dir].remoteCartridge.x, 
               this_cartridge->connections[dir].remoteCartridge.y,
               this_cartridge->connections[dir].remoteCartridge.z,
               this_cartridge->connections[dir].port_no,
               this_cartridge->nodes[this_cartridge->connections[dir].node_index].name);
          } 
       } /* If enabled */
      }/* Cartridges */
}



/************************************************ Map PhysID to SRIO ID **********************************/
void map_physID_to_SRIOID(Topology_t *topology, int active_cartridges[], int num_active_cartridges) {
   int i,j,dir,num_cartridges;
   Cartridge_t *this_cartridge;

   for(i=0;i<topology->num_cartridges;i++) {
     this_cartridge=&topology->cartridges[i]; /* Stored in linear fashion */
     for(j=0;j<num_active_cartridges;j++) {
          if(active_cartridges[j]==this_cartridge->cartridgeID_Phys) {
#ifdef DEBUG_PRINTF
               printf("\n Cluster: Enabling %s",this_cartridge->name);
#endif
               this_cartridge->enabled=TRUE;
               break;
          } 
    }   
  }
}
/******************************************* Add to routing table ************************************************/
int add_to_routing_table_at_index(RoutingTable_t *routingTable,int dest_srioID,int routing_port,int last_routing_table_index) {

    int index=0,hi,retval=ROUTING_TABLE_UNABLE_TO_ADD_AT_INDEX;

   
   /* Check if the Port matches. Since we are adding in an additive fashin, it is okay to 
       check only the upper bound value */
       if(last_routing_table_index==ROUTING_TABLE_INDEX_UNINITIALIZED) {
        /* First time entry to the routing table. Add it in the first entry. */
          index=0;
          routingTable->num_entries=1;
          routingTable->entries[index].lo=routingTable->entries[index].hi=dest_srioID;   
          routingTable->entries[index].port =routing_port;   
          retval=index;

       } else  if((routingTable->entries[last_routing_table_index].port==routing_port) &&  ((routingTable->entries[last_routing_table_index].hi+1)==dest_srioID) ) {
          /* Update the routing table index */
          index = last_routing_table_index;
          routingTable->entries[index].hi=dest_srioID;       
          retval=index;

      } else {
            retval=ROUTING_TABLE_UNABLE_TO_ADD_AT_INDEX;
       }

     return(retval);
}
/******************************************* Add new entry to routing table ************************************************/
int add_new_to_routing_table(RoutingTable_t *routingTable,int dest_srioID,int port) {

    int index;

  
    if(routingTable->num_entries>=8) {
        return ROUTING_TABLE_FULL;
    } 

    index=routingTable->num_entries;
    routingTable->entries[index].lo=dest_srioID;
    routingTable->entries[index].hi=dest_srioID;
    routingTable->entries[index].port=port;

    routingTable->num_entries++;
    return index;
}

/******************************************* Print routing table ************************************************/
void print_routing_table(NodeInfo_t *this_node, RoutingTable_t *routingTable) {

    int i;
    printf("\n Routing Table for %s (# %d )",this_node->name,this_node->srioID);
  
    for(i=0;i<routingTable->num_entries;i++) {
           printf("\n    %s(#%d):   [%d]  (%d  -  %d) --> %d",this_node->name,this_node->srioID,i, routingTable->entries[i].lo,routingTable->entries[i].hi,routingTable->entries[i].port);
   }
}

/*******************************************    Find Routing Tabel Index    *************************************/
int find_routing_table_index(Cartridge_t *this_cartridge,
                                             Cartridge_t *dest_cartridge, 
                                             NodeInfo_t *this_node, 
                                             int dest_srioID,
                                             int last_routing_table_index,
                                             DepthToCartridge_t *depthToDestCartridge, 
                                             DepthToNodes_t DepthToNodes[4][4],
                                             int *routing_table_index) 
{
      /* Find the out_cartridge_port  first, fom the depthToCartridge , and the node_index */
      int outbound_cartridge_port,outbound_dir,outbound_depth,new_index=-1; /* The outbound cartridge port */
      int my_dir_to_gateway_node,my_port_to_gateway_node,gateway_node_index,gateway_node_port,my_outbound_port,dir,orig_outbound_port;
      RoutingTable_t *routingTable;
      NodeInfo_t *gateway_node=NULL;
      DepthToNodes_t *depthToNode;
      int min_cartridge_depth;
      DepthToDest_t  depthToDestPerDir[4]={{MAX_DEPTH,-1}, {MAX_DEPTH,-1},{MAX_DEPTH,-1},{MAX_DEPTH,-1}};
 
      char dir_string[5][15]={"LEFT","RIGHT","UP","DOWN","SAME"}; 


      //printf("\n @find_routing_table_index: Start: For %s, to dest_srioID=%d, to the last routing table index=%d", this_node->name,dest_srioID,last_routing_table_index);
//      fflush(stdout);

      /* Get the minimum depth first */

     if(this_cartridge->cartridgeID_XY==dest_cartridge->cartridgeID_XY) {
        int dest_x,dest_y,dest_z;
           /* Same cartridge. No gateway needed, just the node-to-node could be used */
       get_xyz_from_srio_node_id(dest_srioID,dest_x,dest_y,dest_z)
//      printf("\n              @find_routing_table_index: SameCartridge: my Z=%d, dest Z= %d", this_node->coordinates.z,dest_z);
//      fflush(stdout);
       depthToNode=&(DepthToNodes[this_node->coordinates.z][dest_z]);
       min_cartridge_depth= 0;

     } else {

       outbound_dir = depthToDestCartridge->min_depth_dir; /* Left/Right/Up/Down/Same ..*/
       outbound_depth = depthToDestCartridge->entries_per_dir[outbound_dir].depth; /* Not really used here, unless we choose others */

       gateway_node_index =  this_cartridge->connections[outbound_dir].node_index;
       gateway_node_port =  this_cartridge->connections[outbound_dir].port_no; 
       gateway_node=&this_cartridge->nodes[gateway_node_index];
//       printf("\n find_routing_table_index: %s , DestID = %d  , Gateway Node = %s, port %d",this_node->name, dest_srioID, 																	  gateway_node->name,gateway_node_port);

       /* Find a way to reach the gateway node */
       depthToNode=&(DepthToNodes[this_node->coordinates.z][gateway_node->coordinates.z]);
       min_cartridge_depth=depthToDestCartridge->entries_per_dir[outbound_dir].depth;

     } /* Same Cartridge */

     if(min_cartridge_depth==MAX_DEPTH) {
      printf("\n @find_routing_table_index ERROR! Unreachable destination node #%d@C%s , from %s, Please chech Depth[%s][%s]",dest_srioID,dest_cartridge->name,this_node->name,this_cartridge->name, dest_cartridge->name);
      return(ERR_ROUTING_UNREACHABLE_NODE);
     }
   /* now try adding in range in to the routing table. If a suitable range addition could not be found, try to add in to other ranges. If create one */
       routingTable=   &(this_node->routingTable);

       if(dest_srioID!=this_node->srioID) {
           my_dir_to_gateway_node=depthToNode->dir;
           if(gateway_node!=this_node) {
              orig_outbound_port=my_outbound_port=this_node->connections_per_dir[my_dir_to_gateway_node].port_no;
          } else {
              orig_outbound_port=my_outbound_port=gateway_node_port;
            } 
//         printf("\n find_routing_table_index: %s , DestID = %d  , Outbound Port = %d",this_node->name, dest_srioID,  																	  my_outbound_port);
     
//      printf("\n             @find_routing_table_index: Dir to Gateway is  %s, Port is =%d", dir_string[my_dir_to_gateway_node], my_port_to_gateway_node);
//      fflush(stdout);
       /* Add to the routing table at the previous index */
       new_index=add_to_routing_table_at_index(routingTable,dest_srioID,my_outbound_port,last_routing_table_index);                                    
    
       /* Add to the current ranges. This is okay, because we are adding the IDs in a linear fashion. */
       if(new_index==ROUTING_TABLE_UNABLE_TO_ADD_AT_INDEX) {
                 /* */
          my_outbound_port=orig_outbound_port; 
           if(new_index==ROUTING_TABLE_UNABLE_TO_ADD_AT_INDEX) {
              new_index=add_new_to_routing_table(routingTable,dest_srioID,my_outbound_port);                                    
              if(new_index== ROUTING_TABLE_FULL) {
                printf("\n             ERROR!!!!! ROUTING TABLE FULL !!! find_routing_table_index: %s  Cannot add entry to routing table from #%d  to #%d", this_node->name,this_node->srioID,dest_srioID);
                fflush(stdout);
               return (ERR_ROUTING_TABLE_FULL);              
             }          
         } /* If (new_index==ROUTING_TABLE_UNABLE_TO_ADD_AT_INDEX) */
     }
     } /* dest_srioID!= this_node->srioID) */ else {
            /* Dont add anything to the table. The next invocation to add will force a new entry to the table */
            new_index=last_routing_table_index;
     }
      *routing_table_index=new_index;
      return ROUTING_SUCCESS;
}

/************************* Generate Cartridge Routing Table Entries  ***************************************************************/
int generate_routing_entries (Topology_t *cluster, 
                                                DepthToCartridge_t depthToCartridges[MAX_NUM_CARTRIDGES][MAX_NUM_CARTRIDGES],
                                                int num_cartridges) {

     int catridge_ids_so_far[MAX_NUM_CARTRIDGES],num_cartridges_so_far,dir;
    Cartridge_t *my_cartridge,*dest_cartridge,*adjacent_cartridge,*remote_cartridge;
    int depth[4]={MAX_DEPTH, MAX_DEPTH, MAX_DEPTH, MAX_DEPTH}; /* For each LEFT/RIGHT/UP/DOWN */
    int srio_ID,routing_table_index,last_routing_table_index,dest_srioID,my_cartridge_xy_id,dest_cartridge_xy_id,my_x,my_y,dest_x,dest_y,my_z,dest_z,ret;
    NodeInfo_t *this_node,*dest_node;
   DepthToCartridge_t  *depthToCartridge;
   coordinates_t dest_coordinates;
   int node_no,dest_srioID_base;

   /* Go through the IDs in a progressive manner. This way we could add it to the routing table */
    for(my_y=0;my_y<cluster->num_rows; my_y++) {
       for(my_x=0;my_x<cluster->num_columns; my_x++) {
          
          my_cartridge=&(elem_2darray(cluster->cartridges,my_x,my_y,cluster->num_columns));
          my_cartridge_xy_id=srio_cartridge_id_xy(my_x,my_y);
  
          if(my_cartridge->enabled) {

            for(node_no=0;node_no<4;node_no++) {
               this_node = &my_cartridge->nodes[node_no];
               last_routing_table_index=ROUTING_TABLE_INDEX_UNINITIALIZED;    
              for(dest_y=0;dest_y<cluster->num_rows; dest_y++) {
                 for(dest_x=0;dest_x<cluster->num_columns; dest_x++) {

                    dest_cartridge=&(elem_2darray(cluster->cartridges,dest_x,dest_y,cluster->num_columns));
         
                    if(dest_cartridge->enabled) { 
                      dest_srioID_base = dest_cartridge->nodes[0].srioID;
                      dest_cartridge_xy_id=srio_cartridge_id_xy(dest_x,dest_y);

                     depthToCartridge=&depthToCartridges[my_cartridge_xy_id- XY_ID_BASE][dest_cartridge_xy_id- XY_ID_BASE];
 
                    /* Now, from each node of mine, try to reach (the SRIOID(dest_x,dest_y,dest_z=0-3) */

                            my_z=this_node->coordinates.z;
          
                            /* Find the ports needed to reach out */   

                               for(dest_z=0;dest_z<4;dest_z++) {

                                    
                                 dest_srioID=srio_node_id_xyz(dest_x,dest_y,dest_z);
                                 dest_coordinates.x = dest_x;
                                 dest_coordinates.y = dest_y;
                                 dest_coordinates.z = dest_z;

                                 dest_node=find_Node_by_coordinates(cluster,&dest_coordinates);
                                 if(dest_node->enabled) {

                                 ret = find_routing_table_index(my_cartridge, dest_cartridge, 
                                                                                                          this_node, dest_srioID,
                                                                                                           last_routing_table_index,
                                                                                                           depthToCartridge, 
 													    DepthToNodes, 
                                                                                                            &routing_table_index);
//                                 printf("\n RoutinTableGen: index for node %s to dest srio ID= %d is INDEX=%d",this_node->name,dest_srioID,routing_table_index);
//                                 fflush(stdout);
                                if(ret== ERR_ROUTING_TABLE_FULL) {
                                     printf("\n ERROR , Routing Table Entry could not be addded for Node %s to %d", this_node->name,dest_srioID);
                                     return(ERR_ROUTING_TABLE_FULL);
                                 } else if (ret==ERR_ROUTING_UNREACHABLE_NODE) {
                                     printf("\n ERROR , Unreachable node! %s from %d", this_node->name, dest_srioID);
                                     return(ERR_ROUTING_UNREACHABLE_NODE);
                                 } else {
                                     last_routing_table_index=routing_table_index;
//                                     printf("\n %s : Routing Table Index for SRIOID %d is %d",this_node->name,dest_srioID,routing_table_index);
                                 }
                              } /* Dest node enabled */  
                             } /* dest_z=0;dest_z<3;dest_z++ */
//                             printf("\n  Routing Table Entries for  %s done ",this_node->name);
                      } /*if dest enabled */
                 } /* dest_x */
              }/* dest_y */
#ifdef DEBUG_PRINTF
              print_routing_table(this_node,&this_node->routingTable);
#endif
          

            } /* For nodes 1-4 */

#ifdef DEBUG_PRINTF
              printf("\nRouting Table Index for  %s done ",my_cartridge->name);
#endif

          } /* if enabled */
          
       } /* my_x */
     }/* my_y */
#ifdef DEBUG_PRINTF
        printf("\n %s : Routing Table Index for SRIOID %d is %d",this_node->name,dest_srioID,routing_table_index);
#endif

}



/************************************************************************************************************************************/
/******************************************************* Simulate a walk between two nodes    *******************************************/
/*  The idea here is to initiate a walk from every node to every other node, starting from (0,0,0) to all other nodes.        
/******************************************************* ****************************************************************************/
/************************************************************************************************************************************/
srio_walker_retcode_e simulate_srio_walk(Topology_t *cluster, Cartridge_t *start_cartridge, NodeInfo_t *start_node, int dest_srioID, NodeInfo_t * nodes_so_far[],int *num_nodes_so_far,int *hop_count) {

    NodeInfo_t *walker_node=start_node;
   Cartridge_t * walker_cartridge=start_cartridge;
    NodeInfo_t *dest_node;

    RoutingTable_t *routingTable;
   coordinates_t next_walker_coordinates,dest_coordinates;
    int depth,i, routing_port;
    Rbool next_walker_coordinates_found;
    int hops=0;
    int hops_threshold=MAX_WALKER_HOPS_THRESHOLD;

    /* Find the coordinates of the starting point, so that the node could be accessed easily */
    get_xyz_from_srio_node_id(dest_srioID,dest_coordinates.x,dest_coordinates.y,dest_coordinates.z)
    dest_node=find_Node_by_coordinates(cluster,&dest_coordinates);
    *num_nodes_so_far=0;
    hops=0;

    /* Loop the 'walker' node (starting with the start node)  reaches the destination node, or the  max hop count is reached */
    while((walker_node->srioID != dest_srioID) && (hops < hops_threshold)) {
       
    /* Check for Loops, i.e already visited nodes in the path. Raise an error and exit */
      for(i=0;i<*num_nodes_so_far;i++) {
        if(walker_node->srioID==nodes_so_far[i]->srioID) {
           return WALKER_ERR_FOUND_LOOP;
        }
      }

    /* Register the current node in the visited nodes, to aid loop detection */
       nodes_so_far[*num_nodes_so_far]=walker_node;
       *num_nodes_so_far=*num_nodes_so_far+1;

    /* Check for disabled cartridges on the path */
     if(!(walker_cartridge->enabled)) {
         return WALKER_ERR_CARTRIDGE_DISABLED;
     }

    /* Access the walker's  routing table and find the port to route to */
     routingTable=&(walker_node->routingTable);
     routing_port=-1;

     /*Find the routing port in the routingTable entries */
    for(i=0;i<routingTable->num_entries;i++) {
        if(( dest_srioID >=routingTable->entries[i].lo) && ( dest_srioID <=routingTable->entries[i].hi)) {
                routing_port =routingTable->entries[i].port;
                break;           
        } 
    }

     
     if(routing_port==-1) {
     /* No entries found in the above routing table, hence it is a blocked path. Routing algorithm has failed, return */
        return WALKER_ERR_BLOCKED_PATH;
    }

    /* Now find the adjacent node this 'routing port' is connected to. This could be found by searching the information 
        inside the node in the node->connections[] list. The node->connections[] list is populated at the connection establishment
        time during topology bin creation */

     next_walker_coordinates_found=FALSE;
     for(i=0;i<4;i++) {
         if((walker_node->connections[i].enabled) && (walker_node->connections[i].port_no==routing_port)) {
            next_walker_coordinates=walker_node->connections[i].remoteNode;
            next_walker_coordinates_found=TRUE;
         }       
     }

       if(!next_walker_coordinates_found) { 
        /* Remote coordinates not found! The routing port leads nowhere. Raise alarm and exit */
          return WALKER_ERR_UNINITIALIZED_COORDINATES;

       }
        walker_node=find_Node_by_coordinates(cluster,&next_walker_coordinates);
        walker_cartridge=find_Cartridge_by_coordinates(cluster,&next_walker_coordinates);
        hops++;
       
   } /* While walker_node!=dest_srioID */

   /* Return the hop count */
  *hop_count=hops;
  return WALKER_SUCCESS;
}
/****************************************************************************************************
/*******************************  Simulation of all paths   ***********************************************
/*  The idea here is to initiate a walk from every node to every other node, starting from (0,0,0) to all other nodes.    
/****************************************************************************************************/
void simulate_cluster_paths(Topology_t *cluster)
{
    int catridge_ids_so_far[MAX_NUM_CARTRIDGES],num_cartridges_so_far,dir;
    Cartridge_t *my_cartridge,*dest_cartridge,*adjacent_cartridge,*remote_cartridge;
    int depth[4]={MAX_DEPTH, MAX_DEPTH, MAX_DEPTH, MAX_DEPTH}; /* For each LEFT/RIGHT/UP/DOWN */
    int srio_ID,routing_table_index,dest_srioID,my_cartridge_xy_id,dest_cartridge_xy_id,my_x,my_y,dest_x,dest_y,my_z,dest_z;
    NodeInfo_t *this_node,*dest_node;
    NodeInfo_t *max_hops_src=NULL, *max_hops_dest=NULL;
   DepthToCartridge_t  *depthToCartridge;
   int node_no,dest_srioID_base,hops=0;
   srio_walker_retcode_e walker_ret;
    NodeInfo_t *nodes_so_far[MAX_NUM_CARTRIDGES*MAX_NUM_NODES];
    int num_nodes_so_far,i;
    coordinates_t dest_coordinates;
    int max_hops=0,failed_walks=0,total_walks=0;

   /* Go through the IDs in a progressive manner, starting with the coordinate (0,0,0) to all other coordinates.The SRIO ID, which is a function of coordinates 
        is fetched along the way. Only those cartridges and nodes which are enabled in the cluster are walked by*/
    for(my_y=0;my_y<cluster->num_rows; my_y++) {
       for(my_x=0;my_x<cluster->num_columns; my_x++) {
          /* Find the starting cartridge */          
          my_cartridge=&(elem_2darray(cluster->cartridges,my_x,my_y,cluster->num_columns));
          my_cartridge_xy_id=srio_cartridge_id_xy(my_x,my_y);
  
          /* Walk only if cartridge this is enabled */
          if(my_cartridge->enabled) {

          /* From every node in this cartridge, initiate a walk to every other node in the cluster */
            for(node_no=0;node_no<my_cartridge->num_nodes;node_no++) {
               this_node = &my_cartridge->nodes[node_no];
               if(this_node->enabled) { 

                /* From this node, try to reach every other node with SRIOID(dest_x,dest_y,dest_z) */
                for(dest_y=0;dest_y<cluster->num_rows; dest_y++) {
                 for(dest_x=0;dest_x<cluster->num_columns; dest_x++) {

                    dest_cartridge=&(elem_2darray(cluster->cartridges,dest_x,dest_y,cluster->num_columns));
         
                    if(dest_cartridge->enabled) { 
                      dest_srioID_base = dest_cartridge->nodes[0].srioID;
                      dest_cartridge_xy_id=srio_cartridge_id_xy(dest_x,dest_y);

                           /* Now, from this node of mine, try to reach (the SRIOID(dest_x,dest_y,dest_z) */
                            my_z=this_node->coordinates.z;
          
                               for(dest_z=0;dest_z<4;dest_z++) {

                                 dest_srioID=srio_node_id_xyz(dest_x,dest_y,dest_z);
                                 dest_coordinates.x=dest_x;
                                 dest_coordinates.y=dest_y;
                                 dest_coordinates.z=dest_z;
                                 dest_node=find_Node_by_coordinates(cluster,&dest_coordinates);

                                /* Do not perform walk if the destination node is disabled */
                                  if(dest_node->enabled) {

                                 num_nodes_so_far=0;
                                 hops=0;

                                 printf("\n Walker Simulation: %s to %s",this_node->name,dest_node->name);

                                 walker_ret=simulate_srio_walk(cluster, my_cartridge, this_node,dest_srioID,nodes_so_far,&num_nodes_so_far,&hops);
                                 total_walks++;
                                 switch(walker_ret) {
                                        case WALKER_ERR_FOUND_LOOP:
                                                printf("\n Walker ERROR!! Looped path! on path from %s to %s:", this_node->name,dest_node->name);
                                                break;
                                        case WALKER_ERR_CARTRIDGE_DISABLED:
                                                printf("\n Walker ERROR!! Cartridge Disabled! on path from %s to %s:", this_node->name,dest_node->name);
                                                break;
                                        case WALKER_ERR_BLOCKED_PATH:
                                                printf("\n Walker ERROR!! Blocked path! on path from %s to %s:", this_node->name,dest_node->name);
                                                break;
                                        case WALKER_ERR_UNINITIALIZED_COORDINATES:
                                                printf("\n Walker ERROR!! Uninitialized Coordinates! on path from %s to %s:", this_node->name,dest_node->name);
                                                break;
                                        case WALKER_SUCCESS:
                                                if(max_hops<hops) {
                                                     max_hops=hops;
                                                     max_hops_src=this_node;
                                                     max_hops_dest=dest_node;
                                                }

                                        default:
                                             break;
                                   }

                              if(walker_ret!=WALKER_SUCCESS) {
                                   failed_walks++;
                              } 


                                  if(walker_ret== WALKER_SUCCESS) {
                                          printf(" SUCCESS: %d hops ", hops);
                                   }
                                   /* Print the nodes in the path so far*/
                                   {
                                        for(i=0;i<num_nodes_so_far;i++) {
                                              printf("--> %s ",nodes_so_far[i]->name);
                                        } 
                                    }                                  
                                     if(walker_ret==WALKER_SUCCESS) {
                                             printf("--> %s ",dest_node->name);
                                     }
                              } /* if dest node enabled */      
                            } /* dest_z=0;dest_z<3;dest_z++ */
                      } /*if dest enabled */
                 } /* dest_x */
              }/* dest_y */
             } /* if node enabled */
            } /* For nodes 1-4 */

          } /* if enabled */
          
       } /* my_x */
     }/* my_y */

    printf("\n************************* WALKER SIMULATION RESULTS *********************************");
    printf("\n Max Hop count=%d, from node %s to node %s",max_hops, max_hops_src->name,max_hops_dest->name);
    printf("\n Performed walks = %d", total_walks);
    printf("\n Failed Walks = %d", failed_walks);
    printf("\n***************************************************************************************\n");

}

/***************************************************************************************************//***************************************************************************************************//***************************************************************************************************/
void * srioUtils_LoadFromFile(FILE *topology_file) {

   int bytes_read;
  SRIORunTimeStructures_t *srioRunTimeStructures;

   srioRunTimeStructures = (SRIORunTimeStructures_t *)malloc(sizeof(SRIORunTimeStructures_t));

   if((bytes_read=fread(&(srioRunTimeStructures->cluster),1,sizeof(Topology_t),topology_file))!=sizeof(Topology_t)) {
      printf("\n Error @ read topology %d",sizeof(Topology_t));
      free(srioRunTimeStructures);
      return NULL;
     }

   /* Return the base address */
  return((void *)srioRunTimeStructures);

}

/***************************************************************************************************//***************************************************************************************************//***************************************************************************************************/
int srioUtils_formCluster(void *base,int cartridge_phys_ids[],int num_cartridges) {

      int bytes_read,ret;
      SRIORunTimeStructures_t *srioRunTimeStructures;
      srioRunTimeStructures = (SRIORunTimeStructures_t *)base;

      /* Enabled the cartridges */
      enable_cartridges_by_PhysID(&(srioRunTimeStructures->cluster), cartridge_phys_ids, num_cartridges);

      /* Initialize the Depth to Nodes Table */
      memcpy(&(srioRunTimeStructures->DepthToNodes),DepthToNodes,sizeof(DepthToNodes));

#ifdef DEBUG_PRINTF
       printf("\n Init depth to cartridges ..");
#endif
    /* Now compute the Routing Tables */
       init_depth_to_cartridges(srioRunTimeStructures->DepthToCartridges,MAX_NUM_CARTRIDGES);

#ifdef DEBUG_PRINTF
        printf("\n Generating depth to cartridges ..");
#endif

        generate_depth_to_cartridges(&(srioRunTimeStructures->cluster),srioRunTimeStructures->DepthToCartridges); 

#ifdef DEBUG_PRINTF
        /* Enabling cartridges */  
        print_depth_to_cartridges(&(srioRunTimeStructures->cluster),srioRunTimeStructures->DepthToCartridges,MAX_NUM_CARTRIDGES);
 #endif

       ret=generate_routing_entries ( &(srioRunTimeStructures->cluster), 
                                                srioRunTimeStructures->DepthToCartridges,
                                                MAX_NUM_CARTRIDGES);
      if(ret==ERR_ROUTING_TABLE_FULL) {
         printf("\n ERROR!!! Routing Table Creation:Entries Full ..Exiting..");
         return(SRIO_CLUSTER_ERR_RTABLE_ENTRIES_FULL);
      } else if(ret==ERR_ROUTING_UNREACHABLE_NODE) {
         printf("\n ERROR!!! Routing Table Creation: Unreachable node in the cluster..Exiting..");
         return(SRIO_CLUSTER_ERR_UNREACHABLE_NODE);
      }
   /* Return the base address */
  return(SRIO_CLUSTER_FORMATION_SUCCESS);

}

/*************************************************************************************
********************** Map Phys ID to SRIO id  ******************************************
**************************************************************************************/
int  srioUtils_map_physID_to_SRIOId(void *base,int cartridge,int node,unsigned short *srioID) {

    int srio_port,cols,ret=-1,i,j,physID;
     SRIORunTimeStructures_t *srioRunTimeStructures;
     srioRunTimeStructures = (SRIORunTimeStructures_t *)base;
     Topology_t *topology = &(srioRunTimeStructures->cluster);
     int dir,num_cartridges;
    Cartridge_t *this_cartridge;

    /* Search the topology to match the cartridge physID */ 
   for(i=0;i<topology->num_cartridges;i++) {
     this_cartridge=&topology->cartridges[i]; /* Stored in linear fashion */
//       if(this_cartridge->enabled)
       {
           if(this_cartridge->cartridgeID_Phys==cartridge) {
                for(j=0;j<4;j++) {
                    if(this_cartridge->nodes[j].nodeID_Phys == node) {
                         *srioID=this_cartridge->nodes[j].srioID;
                          return 0;
                    }
                } /* for j: all nodes */
             } /* if cartridgeID_Phys matches */
            } /* if cartridge enabled */
         } /* for all cartridges */  
     return -1;
}

/*************************************************************************************
********************** Map Phys ID to SRIO id  *********************
**************************************************************************************/
int  srioUtils_map_SRIOId_to_physID(void *base,int srioID,int *cartridge,int *node) {
     coordinates_t coordinates;
     SRIORunTimeStructures_t *srioRunTimeStructures;
     srioRunTimeStructures = (SRIORunTimeStructures_t *)base;
     Topology_t *cluster = &(srioRunTimeStructures->cluster);
     Cartridge_t *my_cartridge;
     NodeInfo_t *my_node;
     int ret=-1;

     get_xyz_from_srio_node_id(srioID,coordinates.x,coordinates.y,coordinates.z)
     my_cartridge=&(elem_2darray(cluster->cartridges,coordinates.x,coordinates.y,cluster->num_columns));
     my_node=&(my_cartridge->nodes[coordinates.z]);

     if((my_cartridge->enabled) && (my_node->srioID==srioID)) {
         *cartridge=my_cartridge->cartridgeID_Phys;
         *node=my_node->nodeID_Phys;
          ret=0;
     } 
    return(ret);  
}

/*************************************************************************************
********************** Find my neighbours and fill output_dest_ids[port_no] **********
**************************************************************************************/
int srioUtils_findNeighbours(void *base,int srioID,int output_dest_ids[],int only_active_nodes) {

	     coordinates_t coordinates;
	     coordinates_t remote_coordinates;
	     SRIORunTimeStructures_t *srioRunTimeStructures;
	     srioRunTimeStructures = (SRIORunTimeStructures_t *)base;
	     Topology_t *cluster = &(srioRunTimeStructures->cluster);
	     Cartridge_t *my_cartridge,*remote_cartridge;
	     NodeInfo_t *my_node;
	     int ret=0,i,port;
	     uint8_t remote_id;

	     get_xyz_from_srio_node_id(srioID,coordinates.x,coordinates.y,coordinates.z)
	     my_cartridge=&(elem_2darray(cluster->cartridges,coordinates.x,coordinates.y,cluster->num_columns));
	     my_node=&(my_cartridge->nodes[coordinates.z]);
         for(i=0;i<4;i++) {
	        output_dest_ids[i]=-1; /* Default */
         }

	     if(my_node->srioID==srioID) {
            for(i=0;i<my_node->num_connections;i++) {

                if(!my_node->connections[i].enabled) {
                	continue;
                }

             	port = my_node->connections[i].port_no;
	    	     if(port<0) {
	    	         continue; /* Undefined connection */
	    	     }


	    	     remote_coordinates=my_node->connections[i].remoteNode;
	    	     remote_cartridge=&(elem_2darray(cluster->cartridges,remote_coordinates.x,remote_coordinates.y,cluster->num_columns));
	    	     if(only_active_nodes &&  (!remote_cartridge->enabled) ) {
	    	    	 /* Populate only for active nodes */
	    	    	 continue;
	    	     }

	    	     remote_id=srio_node_id_xyz(remote_coordinates.x,remote_coordinates.y,remote_coordinates.z);
                 //printf("\nsrioUtils_findNeighbours: # %d Connection[%d], Name: %s, Srio# %d on Port %d",srioID,i,my_node->connections[i].node_name,remote_id,port);
	    	     output_dest_ids[port]=(uint8_t)remote_id;
            }
	     }
	    return(ret);
}


/*************************************************************************************
********************** Find routing port   *********************
**************************************************************************************/
int  srioUtils_find_routing_port(const void *base,int srio_id_src, int srio_id_dest) {
     coordinates_t coordinates;
     SRIORunTimeStructures_t *srioRunTimeStructures;
     srioRunTimeStructures = (SRIORunTimeStructures_t *)base;
     Topology_t *cluster = &(srioRunTimeStructures->cluster);
     Cartridge_t *start_cartridge;
     NodeInfo_t *start_node;
     RoutingTable_t *routingTable;
     int ret=-1,i;

     get_xyz_from_srio_node_id(srio_id_src,coordinates.x,coordinates.y,coordinates.z)
#ifdef DEBUG_PRINTF
      printf("\n FindRoutingPort: SRIOID %d, coordinates = (%d,%d,%d)",srio_id_src,coordinates.x,coordinates.y,coordinates.z);
#endif
     start_cartridge=&(elem_2darray(cluster->cartridges,coordinates.x,coordinates.y,cluster->num_columns));
     start_node=&(start_cartridge->nodes[coordinates.z]);

     if(!((start_cartridge->enabled) &&  (start_node->srioID==srio_id_src))) {
         printf("\n ERROR!!! Find routing port: Cannot find starting node!!! Found %s, searched for %d to %d",start_node->name,srio_id_src,srio_id_dest);
         return -1;
     } 

   routingTable = &(start_node->routingTable);  
   for(i=0;i<routingTable->num_entries;i++) {
         
         if( (srio_id_dest >= routingTable->entries[i].lo) && (srio_id_dest <= routingTable->entries[i].hi)) {
           /* Found */
              ret = routingTable->entries[i].port;
              break;
         }
  }
   return ret;  	
}


/****************************************************************************************************
*******************************  Displaying Port Forwarding Regs   ***************************************
****************************************************************************************************/

int srioUtils_displayPortForwardingRegs(void *base, void *srio_base_virtual, int mySrioID) {

   int i;
   uint8_t baseID8Bit;
   uint16_t baseID16Bit;
   uint16_t deviceId, deviceVendorId;
   uint32_t deviceRev;
   CSL_SrioHandle hSrio =   (CSL_SrioHandle )srio_base_virtual;
   /* The base data structure */

    CSL_SRIO_GetDeviceInfo(hSrio,&deviceId,&deviceVendorId,&deviceRev);
    printf("\nSRIO Device Info Device ID=%x, VendorID = %x, Device Rev=%x",deviceId,deviceVendorId,deviceRev);
    CSL_SRIO_GetDeviceIDCSR(hSrio,&baseID8Bit,&baseID16Bit);
    printf("\n More SRIO Device Info 8 bit Device ID=%x,16 bit Device ID=%x \n",baseID8Bit,baseID16Bit);

    printf("\n Current Values of Packet forwarding registers \n");
    printf("\n ------------------------------------------------------------------- \n");

     printf("\n ");  
   for(i=0;i<8;i++) {
     printf("\n   PF_CNTL[%d].RIO_PF_8B_CNTL= 0x%08x , PF_CNTL[%d].RIO_PF_16B_CNTL= 0x%08x  ",i,hSrio->PF_CNTL[i].RIO_PF_8B_CNTL,i,hSrio->PF_CNTL  [i].RIO_PF_16B_CNTL);
	   }
    printf("\n ");
   return 0;
}

/*************************************************************************************
********************** Program Forwarding registers  *********************
**************************************************************************************/
#define RESET_AT_ENDPOINTS 1
int  srioUtils_progFwdingRegs(void *base,  void *srio_base_virtual, int mySrioID) {

 int srio_port,cols,ret=-1,physID=-1,x,y,i;
 CSL_SrioHandle hSrio;
  int num_rows;
  int num_columns;
  int outBoundID_hi, outBoundID_lo,forwardingPort;
  Rbool found=FALSE;
     coordinates_t coordinates;
     SRIORunTimeStructures_t *srioRunTimeStructures;
     srioRunTimeStructures = (SRIORunTimeStructures_t *)base;
     Topology_t *cluster = &(srioRunTimeStructures->cluster);
     Cartridge_t *my_cartridge;
     NodeInfo_t *my_node;
     RoutingTable_t *routingTable=NULL;
     int first_port=0;
    Rbool end_point;
     get_xyz_from_srio_node_id(mySrioID,coordinates.x,coordinates.y,coordinates.z)
     my_cartridge=&(elem_2darray(cluster->cartridges,coordinates.x,coordinates.y,cluster->num_columns));
     my_node=&(my_cartridge->nodes[coordinates.z]);

     if(!(my_cartridge->enabled && my_node->srioID==mySrioID)) {
         printf("\n Find routing port: Cannot find starting node!!!");
         return -1 ;
     } 
   routingTable=&(my_node->routingTable);
                  
    
#ifdef RESET_AT_ENDPOINTS
     /* Check if this is an end point */
       end_point=TRUE; /* Assuming this is an end point */
       first_port=routingTable->entries[0].port;
       for(i=1;i<routingTable->num_entries;i++) {    
           if(routingTable->entries[i].port != first_port) {
             end_point=FALSE;  /* There are traffic to various directions , and this is NOT an end node */
             break;
          }
      }
#else
end_point=FALSE;
#endif
       if(end_point==FALSE) {
       /* Now that the routing table element is found, program the registers */
        hSrio =   (CSL_SrioHandle )srio_base_virtual;
       for(i=0;i<routingTable->num_entries;i++) {    

            outBoundID_lo=routingTable->entries[i].lo;
            outBoundID_hi=routingTable->entries[i].hi;
            forwardingPort=routingTable->entries[i].port;

           hSrio->PF_CNTL[i].RIO_PF_8B_CNTL=(forwardingPort<<CSL_SRIO_RIO_PF_8B_CNTL_OUT_PORT_SHIFT) |
                                                          ((outBoundID_hi &0xff) <<CSL_SRIO_RIO_PF_8B_CNTL_DEVID_8B_UP_SHIFT)|
                                                          ((outBoundID_lo &0xff) << CSL_SRIO_RIO_PF_8B_CNTL_DEVID_8B_LO_SHIFT);

           hSrio->PF_CNTL[i].RIO_PF_16B_CNTL=(outBoundID_hi<<CSL_SRIO_RIO_PF_16B_CNTL_DEVID_16B_UP_SHIFT) |
                                                           (outBoundID_lo<<CSL_SRIO_RIO_PF_16B_CNTL_DEVID_16B_LO_SHIFT);

      }
  }
   return ;
}

/*************************************************************************************
********************** CUSTOM Program Forwarding registers  *********************
**************************************************************************************/
int  srioUtils_progFwdingRegsCustom(void *base,  void *srio_base_virtual, int mySrioID,RoutingTable_t *routingTable) {

 int i;
 CSL_SrioHandle hSrio;
  int outBoundID_hi, outBoundID_lo,forwardingPort;

    /* Now that the routing table element is found, program the registers */
   if(routingTable!=NULL) {
        hSrio =   (CSL_SrioHandle )srio_base_virtual;
       for(i=0;i<routingTable->num_entries;i++) {    

            outBoundID_lo=routingTable->entries[i].lo;
            outBoundID_hi=routingTable->entries[i].hi;
            forwardingPort=routingTable->entries[i].port;

           hSrio->PF_CNTL[i].RIO_PF_8B_CNTL=(forwardingPort<<CSL_SRIO_RIO_PF_8B_CNTL_OUT_PORT_SHIFT) |
                                                          ((outBoundID_hi &0xff) <<CSL_SRIO_RIO_PF_8B_CNTL_DEVID_8B_UP_SHIFT)|
                                                          ((outBoundID_lo &0xff) << CSL_SRIO_RIO_PF_8B_CNTL_DEVID_8B_LO_SHIFT);

           hSrio->PF_CNTL[i].RIO_PF_16B_CNTL=(outBoundID_hi<<CSL_SRIO_RIO_PF_16B_CNTL_DEVID_16B_UP_SHIFT) |
                                                           (outBoundID_lo<<CSL_SRIO_RIO_PF_16B_CNTL_DEVID_16B_LO_SHIFT);

     }
  } else {
     return 0;
  }
  return;
}
/*************************************************************************************
********************** Reset Forwarding registers  ****************************************
**************************************************************************************/
int  srioUtils_resetFwdingRegs(void *base,  void *srio_base_virtual, int mySrioID) {

 int srio_port,cols,ret=-1,physID=-1,x,y,i;
 CSL_SrioHandle hSrio;
  int num_rows;
  int num_columns;
  int outBoundID_hi, outBoundID_lo,forwardingPort;


    /* Now that the routing table element is found, program the registers */
     hSrio =   (CSL_SrioHandle )srio_base_virtual;
     for(i=0;i<8;i++) {    

            outBoundID_lo=0xffff;
            outBoundID_hi=0xffff;
            forwardingPort=0;

           hSrio->PF_CNTL[i].RIO_PF_8B_CNTL=(forwardingPort<<CSL_SRIO_RIO_PF_8B_CNTL_OUT_PORT_SHIFT) |
                                                          ((outBoundID_hi &0xff) <<CSL_SRIO_RIO_PF_8B_CNTL_DEVID_8B_UP_SHIFT)|
                                                          ((outBoundID_lo &0xff) << CSL_SRIO_RIO_PF_8B_CNTL_DEVID_8B_LO_SHIFT);

           hSrio->PF_CNTL[i].RIO_PF_16B_CNTL=(outBoundID_hi<<CSL_SRIO_RIO_PF_16B_CNTL_DEVID_16B_UP_SHIFT) |
                                                           (outBoundID_lo<<CSL_SRIO_RIO_PF_16B_CNTL_DEVID_16B_LO_SHIFT);
    }
 
   return 0;
}

/******************************************* Print the cluster ***************************************/
int srioUtils_displayCluster(void *base) {
     SRIORunTimeStructures_t *srioRunTimeStructures = (SRIORunTimeStructures_t *)base;
     Topology_t *cluster = &(srioRunTimeStructures->cluster);
      print_cluster(cluster);

      print_depth_to_cartridges(cluster,srioRunTimeStructures->DepthToCartridges,MAX_NUM_CARTRIDGES);
       return 0;
}

/******************************************* Reset the cluster ***************************************/
int srioUtils_ResetCluster(void *base) {
     SRIORunTimeStructures_t *srioRunTimeStructures = (SRIORunTimeStructures_t *)base;
     Topology_t *cluster = &(srioRunTimeStructures->cluster);
     reset_cluster(cluster);
    return 0;
}

