#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include "CJSON/cJSON.h"
#include "routingTableGen.h"


/* Pre assigned XY for cartridges */
CartridgeXYEntries_t moonshotXYEntries[NUM_MOONSHOT_ROWS*NUM_MOONSHOT_COLUMNS] = {
{"c1"},{"c4"},{"c7"},{"c10"},{"c13"},{"c16"},{"c28"},{"c31"},{"c34"},{"c37"},{"c40"},{"c43"},{"c19"},{"c22"},{"c25"}, /* Y=0 */
{"c2"},{"c5"},{"c8"},{"c11"},{"c14"},{"c17"},{"c29"},{"c32"},{"c35"},{"c38"},{"c41"},{"c44"},{"c20"},{"c23"},{"c26"},  /* Y=1 */
{"c3"},{"c6"},{"c9"},{"c12"},{"c15"},{"c18"},{"c30"},{"c33"},{"c36"},{"c39"},{"c42"},{"c45"},{"c21"},{"c24"},{"c27"}  /* Y=2 */ 
};

CartridgeXY_t moonshotXY= {
   NUM_MOONSHOT_ROWS,
   NUM_MOONSHOT_COLUMNS, 
   moonshotXYEntries
};
CartridgeXYEntries_t evmXYEntries[NUM_EVM_ROWS*NUM_EVM_COLUMNS] = {
{"c1"} /* Y=0 */
};

CartridgeXY_t evmXY= {
   NUM_EVM_ROWS,
   NUM_EVM_COLUMNS, 
   evmXYEntries
};


/* Pre assigned XY for cartridges */
CartridgeXYEntries_t ponkXYEntries[NUM_PONK_ROWS*NUM_PONK_COLUMNS] = {
{"c1"}, {"dummy"},/* Y=0 */
{"c2"},{"c5"},  /* Y=1 */
{"c3"},{"dummy"} /* Y=2 */ 
};

CartridgeXY_t ponkshotXY= {
   NUM_PONK_ROWS,
   NUM_PONK_COLUMNS, 
   ponkXYEntries
};


/************************************************ Returns direction ********************************/
int get_xy_dir(int s_x, int s_y, int d_x, int d_y, int x_max, int y_max) 
{
   int dir;
   if( (((s_x +1)% x_max) == d_x) && (s_y == d_y) ) {
        dir=RIGHT; 
   } else if( (((d_x +1)% x_max) == s_x) && (s_y == d_y) ) {
        dir=LEFT; 
   } else if( (((s_y +1)% y_max) == d_y) && (s_x == d_x) ) {
        dir=UP; 
   } else if( (((d_y +1)% y_max) == s_y) && (s_x == d_x) ) {
        dir=DOWN; 
   } else {
        dir=SAME; 
   }
   return dir;
}
/************************************************ Parse CJSON ********************************/
int cjson_topology_parse(char *file,Topology_t *topology)
{
	FILE *fpr;
	char *config_buf = NULL;
	struct stat buf;
	int flen,num_cartridges,num_nodes,num_connections;
	int res, cartridge_no,node_no,connection_no,cartridgeID_Phys,nodeID_Phys;
	cJSON *json = NULL;
        cJSON *cjson_cartridges_array,*cjson_cartridge,*cjson_nodes_array,*cjson_node,*cjson_connections_array,*cjson_connection,*cjson_attribute;
        char ports_attribute_str[4][10]= {"port0","port1","port2","port3"};
        Cartridge_t *this_cartridge; 
        NodeInfo_t *this_node;

	int retval = -1;

	fpr = fopen(file, "r");

        if (!fpr){

		printf("\nError opening file %s ",
				file);
		goto close_n_exit;
	}

	fstat(fileno(fpr), &buf);
	flen = buf.st_size;
        printf("\n Allocated config_buf of length %d",flen+1);
        fflush(stdout);
	config_buf = (char *) malloc(flen + 1);
        printf("\n Allocated config_buf of length %d",flen+1);
        fflush(stdout);
	if (!config_buf) {

		printf("\nError in allocating cfg buffer");
		goto close_n_exit;

	}

	config_buf[flen] = 0;

	res = fread (config_buf, 1, flen, fpr);

	if (res != flen) {

		printf("\nError in reading cfg file..");
                fflush(stdout);
		goto close_n_exit;

	}

        printf("\n First Json parse begins");
        fflush(stdout);

	json = cJSON_Parse(config_buf);
        printf("\n First Json parse complete, got json= %p",json);
        fflush(stdout);

	if (!json) {

		printf("\nJSON parser error before: [%s]",cJSON_GetErrorPtr());
                fflush(stdout);
		goto close_n_exit;

	}

        cjson_cartridges_array = cJSON_GetObjectItem(json,"cartridges");
	num_cartridges = cJSON_GetArraySize(cjson_cartridges_array);
        topology->num_cartridges=num_cartridges;

       printf("\n JSON Parse: Number of cartridges: is %d", num_cartridges);
       fflush(stdout);
      for(cartridge_no=0;cartridge_no<num_cartridges;cartridge_no++) {
                        this_cartridge=&topology->cartridges[cartridge_no];
           		cjson_cartridge = cJSON_GetArrayItem(cjson_cartridges_array, cartridge_no);
                        cjson_attribute = cJSON_GetObjectItem(cjson_cartridge, "name");
                        strcpy(this_cartridge->name,cjson_attribute->valuestring);
                         sscanf(this_cartridge->name,"c%d",&cartridgeID_Phys);     
                        this_cartridge->cartridgeID_Phys=cartridgeID_Phys;/* By default. Forming a cluster will actually enable it */
                        this_cartridge->enabled=FALSE;/* By default. Forming a cluster will actually enable it */

                        cjson_nodes_array    = cJSON_GetObjectItem(cjson_cartridge, "nodes");
             	        num_nodes = cJSON_GetArraySize(cjson_nodes_array);
                       this_cartridge->num_nodes=num_nodes;
                       for(node_no=0;node_no<num_nodes;node_no++) {
                        /* Node Information */
           		cjson_node = cJSON_GetArrayItem(cjson_nodes_array, node_no);
                        cjson_attribute = cJSON_GetObjectItem(cjson_node, "name");

                        this_node=&this_cartridge->nodes[node_no];
                        this_node->enabled=TRUE;
                         strcpy(this_node->name,cjson_attribute->valuestring);
                         strcpy(this_node->cartridgeName, this_cartridge->name);
                          sscanf(this_node->name,"c%dn%d",&cartridgeID_Phys,&nodeID_Phys);     
                         this_node->nodeID_Phys=nodeID_Phys;
                        cjson_connections_array    = cJSON_GetObjectItem(cjson_node, "connections");
             	        num_connections = cJSON_GetArraySize(cjson_connections_array);
                        this_node->num_connections=num_connections;
                        printf("\n %s num connections = %d",this_node->name, this_node->num_connections);
                        for(connection_no=0;connection_no<num_connections;connection_no++) {
          		   cjson_connection = cJSON_GetArrayItem(cjson_connections_array, connection_no);
                           cjson_attribute = cJSON_GetObjectItem(cjson_connection, ports_attribute_str[connection_no]);
                           strcpy(this_node->port_connections[connection_no],cjson_attribute->valuestring);
                       }
               }
      }

	retval = 0;
        if(num_cartridges==NUM_MOONSHOT_CARTRIDGES) {
           topology->num_columns=NUM_MOONSHOT_COLUMNS;
           topology->num_rows=NUM_MOONSHOT_ROWS;
       } else if(num_cartridges==NUM_EVM_CARTRIDGES) {
           topology->num_columns=NUM_EVM_COLUMNS;
           topology->num_rows=NUM_EVM_ROWS;
        } else if (num_cartridges==4) {
           topology->num_columns=NUM_PONK_COLUMNS;
           topology->num_rows=NUM_PONK_ROWS;
        } else {
             printf("\n %s Unsupported Topology!!!!! Exiting\n");
             retval=-1;
        }
close_n_exit:
        if(json) cJSON_Delete(json);
	if (config_buf) free(config_buf);
	if (fpr) fclose(fpr);
	return retval;

}
/************************************************ PRINT TOPOLOGY ***********************************************/
int print_topology(Topology_t *topology)
{
      int port_connections_no,node_connection_no,cartridge_connection_no,node_no,cartridge_no,port_no;
     char *c_name,*node_name;
       Cartridge_t *this_cartridge; 
        NodeInfo_t *this_node;
       char direction[4][15]={"Left","Right","Up","Down"};

      printf("\n ------TOPOLOGY DISPLAY---------\n");
      printf("\n Num Rows= %d, Num Columns = %d",topology->num_rows, topology->num_columns);
    
    for(cartridge_no=0;cartridge_no<topology->num_cartridges; cartridge_no++) {
          /* Display Cartridge Info */
          c_name=topology->cartridges[cartridge_no].name;
           this_cartridge=&topology->cartridges[cartridge_no];
 
           printf("\n   %s:    ENABLED= %d",c_name, this_cartridge->enabled); 
           printf("\n   %s:    Name: %s    ",c_name,this_cartridge->name);
           printf("\n   %s:    XY =(%d,%d)     ",c_name,this_cartridge->coordinates.x,this_cartridge->coordinates.y);
           printf("\n   %s:    Number of Nodes=%d",c_name,this_cartridge->num_nodes);

          /* Display Nodes */
         for(node_no=0;node_no< this_cartridge->num_nodes; node_no++) {
             this_node=&this_cartridge->nodes[node_no];
             node_name=this_node->name; 
             printf("\n   %s:    XYZ =(%d,%d,%d)   ",this_node->name,this_node->coordinates.x,this_node->coordinates.y,this_node->coordinates.z);
             printf("\n   %s:    Connections Initialized = %d",node_name,this_node->connections_per_dir_initialized);

             for(port_connections_no=0;port_connections_no<4;port_connections_no++) {  
                 printf("\n   %s:  Port[%d] --> %s ",node_name, port_connections_no,this_node->port_connections[port_connections_no]);
            }

           /* Node connections */
          for(node_connection_no=0;node_connection_no< 4; node_connection_no++) {
              printf("\n   %s:  Port No %d connected to %s Node at  (X=%d,Y=%d,Z=%d)",
                                                this_node->name, 
                                                this_node->connections_per_dir[node_connection_no].port_no,
                                                this_node->connections_per_dir[node_connection_no].node_name, 
                                                this_node->connections_per_dir[node_connection_no].remoteNode.x,
                                                this_node->connections_per_dir[node_connection_no].remoteNode.y,
                                                this_node->connections_per_dir[node_connection_no].remoteNode.z);

            } 
        }


#if 0        /* Cartridge connections */
        printf("\n Cartridge Connections");
        for(cartridge_connection_no=0;cartridge_connection_no< this_cartridge->num_connections; cartridge_connection_no++) {
                

                printf("\n %s connected on %s to %s  , through port %d, via node %s  ",this_cartridge->name,direction[cartridge_connection_no],
               this_cartridge->connections[cartridge_connection_no].cartridge_name,
               this_cartridge->connections[cartridge_connection_no].remoteCartridge_index,
               this_cartridge->connections[cartridge_connection_no].node_index,
               this_cartridge->connections[cartridge_connection_no].port_no);
           } 
#endif

      }/* Cartridges */

}

/**************************************************************************************************************/
int print_topology_connected(Topology_t *topology)
{
      int port_connections_no,node_connection_no,cartridge_connection_no,node_no,cartridge_no,port_no,dir;
     char *c_name,*node_name;
       Cartridge_t *this_cartridge; 
        NodeInfo_t *this_node;
       char direction[4][15]={"Left","Right","Up","Down"};

      printf("\n ------TOPOLOGY DISPLAY---------\n");
      printf("\n Num Rows= %d, Num Columns = %d",topology->num_rows, topology->num_columns);
    
    for(cartridge_no=0;cartridge_no<topology->num_cartridges; cartridge_no++) {
          /* Display Cartridge Info */
          c_name=topology->cartridges[cartridge_no].name;
           this_cartridge=&topology->cartridges[cartridge_no];
 
           printf("\n\n   %s:    ENABLED= %d",c_name, this_cartridge->enabled); 
           printf("\n   %s:    Name: %s    ",c_name,this_cartridge->name);
           printf("\n   %s:    XY =(%d,%d)     ",c_name,this_cartridge->coordinates.x,this_cartridge->coordinates.y);
           printf("\n   %s:    Number of Nodes=%d",c_name,this_cartridge->num_nodes);

          /* Display Nodes */
         for(node_no=0;node_no< this_cartridge->num_nodes; node_no++) {
             this_node=&this_cartridge->nodes[node_no];
             node_name=this_node->name; 
             printf("\n\n   %s:    XYZ =(%d,%d,%d)   ",this_node->name,this_node->coordinates.x,this_node->coordinates.y,this_node->coordinates.z);
             printf("\n   %s:    Connections Initialized = %d",node_name,this_node->connections_per_dir_initialized);

             for(port_connections_no=0;port_connections_no<4;port_connections_no++) {  
                 printf("\n   %s:  Port[%d] --> %s ",node_name, port_connections_no,this_node->port_connections[port_connections_no]);
            }

           /* Node connections */
               printf("\n Find node connections by coordinates ..");
          for(node_connection_no=0;node_connection_no< 4; node_connection_no++) {
             if(this_node->connections_per_dir[node_connection_no].enabled) {
                  NodeInfo_t *connected_node =find_Node_by_coordinates(topology, &(this_node->connections_per_dir[node_connection_no].remoteNode));
                   printf("\n   %s:  Port No %d connected to %s",
                                                this_node->name, 
                                                this_node->connections_per_dir[node_connection_no].port_no,
                                                connected_node->name);
               } else {
                   printf("\n   %s:  Port No %d connected to NONE",
                                                this_node->name, 
                                                this_node->connections_per_dir[node_connection_no].port_no);
        
               }
            } 
        }

        /* Cartridge connections */
         /* Cartridge connections */
        printf("\n Cartridge Connections");
        for(dir=0;dir< 4; dir++) {
                Cartridge_t *connected_cartridge =find_Cartridge_by_coordinates(topology, &(this_cartridge->connections[dir].remoteCartridge));
                printf("\n %s Connected on %s  to %s ( %d,%d,%d), through port %d, via node %s  ",this_cartridge->name,
               direction[dir],
               connected_cartridge->name,
               this_cartridge->connections[dir].remoteCartridge.x,
               this_cartridge->connections[dir].remoteCartridge.y,
               this_cartridge->connections[dir].remoteCartridge.z,
               this_cartridge->connections[dir].port_no,
               this_cartridge->nodes[this_cartridge->connections[dir].node_index].name);
          } 
      }/* Cartridges */
}


/************************************************************************************************/
/************************************************ Assign X-Y ***********************************************/
int topology_assign_xy (Topology_t *topology, CartridgeXY_t *pre_assigned_xy)
{
      int port_connections_no,node_connection_no,cartridge_connection_no,node_no,cartridge_no,port_no,x,y;
     char *c_name,*node_name;
       Cartridge_t *this_cartridge; 
       CartridgeXYEntries_t *xyEntries_elem;
        NodeInfo_t *this_node;
       Rbool found;

      printf("\n ------TOPOLOGY DISPLAY---------\n");
      printf("\n Num Rows= %d, Num Columns = %d",topology->num_rows, topology->num_columns);
    
    for(cartridge_no=0;cartridge_no<topology->num_cartridges; cartridge_no++) {
          /* Display Cartridge Info */
          c_name=topology->cartridges[cartridge_no].name;
           this_cartridge=&topology->cartridges[cartridge_no];
           found=FALSE;
        /* Search for this_cartridge->name in the pre-assigned_xy_table */
           for(y=0;y<pre_assigned_xy->num_rows;y++) {
               for(x=0;x<pre_assigned_xy->num_columns;x++) {
                      xyEntries_elem=&(elem_2darray(pre_assigned_xy->entries,x,y,pre_assigned_xy->num_columns));
                      if(!strcmp(this_cartridge->name,xyEntries_elem->name)) {
                             found=TRUE;
                             break;
                      }
               } 
                   if(found) 
                      break;
           }
             
           if(!found) {
                    printf("\n X-Y Entry Not found for %s",this_cartridge->name);
                    return -1;
           }
             
          this_cartridge->coordinates.x=x;
          this_cartridge->coordinates.y=y;
          this_cartridge->cartridgeID_XY=srio_cartridge_id_xy(x,y);
          this_cartridge->xy_assigned=TRUE;
          
          /* Assign for all the nodes */
         for(node_no=0;node_no< this_cartridge->num_nodes; node_no++) {
             this_node=&this_cartridge->nodes[node_no];

             this_node->coordinates.x=x;
             this_node->coordinates.y=y;
             this_node->coordinates.z=node_no;
             this_node->srioID=srio_node_id_xyz(x,y,node_no);
             this_node->xyz_assigned=TRUE;

            }
      }/* Cartridges */
}

/************************************************ Sort based on X,Y for cartridges ***********************************************/
int topology_sort (Topology_t *unsorted_topology, Topology_t *sorted_topology)
{
      int port_connections_no,node_connection_no,cartridge_connection_no,node_no,cartridge_no,port_no,x,y;
     char *c_name,*node_name;
       Cartridge_t *this_cartridge, *sorted_cartridge_elem; 
       Rbool found;

         sorted_topology->num_columns=unsorted_topology->num_columns;
        sorted_topology->num_rows=unsorted_topology->num_rows;
        sorted_topology->num_cartridges= unsorted_topology->num_cartridges;
    for(cartridge_no=0;cartridge_no<unsorted_topology->num_cartridges; cartridge_no++) {
          /* Display Cartridge Info */
           this_cartridge=&unsorted_topology->cartridges[cartridge_no];
            x=this_cartridge->coordinates.x;
            y=this_cartridge->coordinates.y;
            sorted_cartridge_elem=&(elem_2darray(sorted_topology->cartridges,x,y,unsorted_topology->num_columns));

           *sorted_cartridge_elem = *this_cartridge;
      }/* Cartridges */
}

/************************************************ Establish connections within a sorted topology ********************************/
int topology_establish_connection (Topology_t *topology)
{

      int cartridge_connection_no,node_no,cartridge_no,port_no,x,y,port_connection_no,dir;
      char *c_name,*node_name;
      Cartridge_t *this_cartridge; 
      CartridgeXYEntries_t *xyEntries_elem;
      NodeInfo_t *this_node;
      Rbool found;

     for(cartridge_no=0;cartridge_no<topology->num_cartridges; cartridge_no++) {
          /* Display Cartridge Info */
           this_cartridge=&topology->cartridges[cartridge_no];
           found=FALSE;
#ifdef DEBUG_PRINTF
            printf("\n @Establish_Connection: cartridge: %s", this_cartridge->name);
#endif
         /* Set up the node connections */
        /* Display Nodes */
         for(node_no=0;node_no< this_cartridge->num_nodes; node_no++) {
             this_node=&this_cartridge->nodes[node_no];
#ifdef DEBUG_PRINTF
            printf("\n @Establish_Connection: Nodes[%d]: %s with %d connections", node_no,this_node->name, this_node->num_connections);
#endif
            for(port_connection_no=0; port_connection_no<this_node->num_connections; port_connection_no++) {

#ifdef DEBUG_PRINTF
           printf("\n @Establish_Connection: Nodes Connection[%d]: %s", port_connection_no ,this_node->port_connections[port_connection_no]);
           fflush(stdout);
#endif
               if(strcmp(this_node->port_connections[port_connection_no],"NONE")) {
                   coordinates_t coordinates;
                   if((found=find_node_by_name(topology, this_node->port_connections[port_connection_no],&coordinates))==TRUE) {
                        /* Make connection */
                           
                                 this_node->connections[port_connection_no].remoteNode=coordinates;
                                 this_node->connections[port_connection_no].enabled=TRUE;
                                 this_node->connections[port_connection_no].port_no=port_connection_no; 


                        /* Make a connection in the node->connection[LEFT/RIGHT/UP/DOWN] */                         
                        if( (this_node->coordinates.x== coordinates.x) &&  (this_node->coordinates.y== coordinates.y)) {
                            /* Same cartridge. So establish, LEFT/RIGHT connection within. Use Z values, with Y=0.*/
                             if((dir= get_xy_dir(this_node->coordinates.z,
                                        0,
                                      coordinates.z,
                                       0,
                                      4,1))!=SAME) {

                                   this_node->connections_per_dir[dir].remoteNode=coordinates;
                                   this_node->connections_per_dir[dir].enabled=TRUE;
                                  this_node->connections_per_dir[dir].port_no=port_connection_no;
#ifndef ENABLE_LOOP_1_4
                           { 
                                /* Cut the 1_4 loop */
                               if(  (this_node->coordinates.z==0 && dir ==LEFT) || 
                                    (this_node->coordinates.z==3 && dir ==RIGHT))    { /* Node 1 or Node 4*/ 
                                    printf("\n Disconnecting Loops !!");
                                    this_node->connections_per_dir[dir].enabled=FALSE;
                                    this_node->connections[port_connection_no].enabled=FALSE;
                               }
                         }
#endif
                             } else {
                                printf("\n Same Node to Node connection! ERROR!!");
                             }
                         } else {

                         /* Make entry in the cartridge connections UP/DOWN/SOUTH/NORTH */
                          if((dir=get_xy_dir(this_cartridge->coordinates.x,
                                      this_cartridge->coordinates.y,
                                      coordinates.x,
                                      coordinates.y,
                                      topology->num_columns,topology->num_rows))!=SAME) {
                                      /* Add to the cartridge connections */
                                    coordinates.z=0; /* For cartridges, this is irrelevant */ 
                                   this_cartridge->connections[dir].remoteCartridge=coordinates;
                                   this_cartridge->connections[dir].node_index=node_no;
                                   this_cartridge->connections[dir].port_no=port_connection_no;
                                   this_cartridge->connections[dir].enabled=TRUE;
                           } 
                       } /* not same catridge */
                      } else {
                     printf("\n ERROR !! Connecting node could %s not be found from %s,",this_node->port_connections[port_connection_no],this_node->name);
                 } 
             
              }  else {
                    this_node->connections_per_dir[port_connection_no].enabled=FALSE;
                    this_node->connections[port_connection_no].enabled=FALSE;
              } 
         } /* port connections */     
     } /* Nodes */
  } /* Cartridges */

}






