/*  (C) Copyright 2012-2014, Texas Instruments, Inc.
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h> 
#include <sys/wait.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include "fw_test.h"
#include "fw_mem_allocator.h"
#include "srio_test.h"
#include "srioPlatCfg.h"


uint32_t errorCount = 0;

void  *topLevelTest (void *args);
/*************************************************************/
/** Development APIs: to be removed once we have the other 7 */
/*************************************************************/
extern void Srio_Type11_usageTsk(Cppi_Handle cppiHnd, void *args, int *nodeId, int *nodeRole, 
                              remoteInfo_t *remoteInfo, Srio_PDSPRouterCfg_t *PDSProuterConfig, int dest_ids_port[]);
extern void Srio_Type11_multiNodeTestTask(int my_id, Role_e Role, remoteInfo_t *remoteInfo, Srio_PDSPRouterCfg_t *PDSProuterConfig, int dest_id[]);

/*************************************************************/

typedef pthread_t task_handle;

#define DEFAULT_STACK_SIZE  0x8000

/* look up value by name */
uint32_t fw_name_lookup (const char *name)
{
    extern uint32_t ex_rm_name_lookup (const char *name);
    /* Bind to example's RM binding to avoid duplicating DSP code */
    return ex_rm_name_lookup (name);
}

/* set name to value */
 void fw_name_set (const char *name, uint32_t val)
{
    extern void ex_rm_name_set (const char *name, uint32_t val);
    /* Bind to example's RM binding to avoid duplicating DSP code */
    ex_rm_name_set (name, val);
}

/* delete name */
 void fw_name_del (const char *name)
{
    extern void ex_rm_name_del (const char *name);
    /* Bind to example's RM binding to avoid duplicating DSP code */
    ex_rm_name_del (name);
}

/** ============================================================================
 *   @n@b task_create
 *
 *   @b Description
 *   @n Create thread to run the test program
 *
 *   @param[in]  
 *   @n None
 * 
 *   @return    int32_t
 *              -1      -   Error
 *              0       -   Success
 * =============================================================================
 */
int fw_task_create ( void *(start_routine)(void*), void* args, void* handle)
{
    int max_priority, err;
    pthread_t thread;
    pthread_attr_t attr;
    struct sched_param param;

    max_priority = sched_get_priority_max(SCHED_FIFO);
    err = pthread_attr_init(&attr);
    if (err) {
        printf("pthread_attr_init failed: (%s)\n", strerror(err));
        return err;
    }
    err = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    if (err) {
        printf("pthread_attr_setdetachstate failed: (%s)\n", strerror(err));
        return err;
    }
    err = pthread_attr_setstacksize(&attr, DEFAULT_STACK_SIZE);
    if (err) {
        printf("pthread_attr_setstacksize failed: (%s)\n", strerror(err));
        return err;
    }
    err = pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED);
    if (err) {
        printf("pthread_attr_setinheritsched failed: (%s)\n", strerror(err));
        return err;
    }
    err = pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
    if (err) {
        printf("pthread_attr_setschedpolicy failed: (%s)\n", strerror(err));
        return err;
    }
    memset(&param, 0, sizeof(param));
    param.sched_priority = max_priority;
    err = pthread_attr_setschedparam(&attr, &param);
    if (err) {
        printf("pthread_attr_setschedparam failed: (%s)\n", strerror(err));
        return err;
    }
    if (err) return err;
    err = pthread_create(&thread, &attr, start_routine, args);
    if (err) {
        printf("pthread_create failed: (%s)\n", strerror(err));
        return err;
    }
    if (err) return err;
    *(pthread_t*)handle = thread;
    return 0;
}

/** ============================================================================
 *   @n@b task_wait
 *
 *   @b Description
 *   @n Wait for Task completion
 * 
 *   @return    void
 * =============================================================================
 */
void fw_task_wait (void *handle)
{
    pthread_join(*((pthread_t*)handle), NULL);
    return;
}

uint32_t l2_global_address(uint32_t x)
{
    return x;
}

/** ============================================================================
 *   @n@b main
 *
 *   @b Description
 *   @n test application main
 * 
 *   @return    int
 * =============================================================================
 */
#ifdef STANDALONE
int main(int argc, char *argv[]) {
#else
int TYPE11_main(int argc, char *argv[]) {
#endif
    task_handle test_th;
    int status;
    int child = 0;
    int parent = 0;
    int i;
    int retVal = 0;
    off_t pCMemBase;

    coreNum = 0;
    parent = 1;
    Osal_srioLog ("pid=%d\n", getpid());

    if ((status = fw_task_create(topLevelTest, argv, &test_th))) {
        printf("ERROR: \"Top Level Test\" task-create failed (%d)\n", status);
        return (-1);
    }
    fw_task_wait(&test_th);

    return retVal;
}

/** ============================================================================
 *   @n@b topLevelTest
 *
 *   @b Description
 *   @n Routine running as a separate thread
 * 
 *   @return    int
 * =============================================================================
 */
void *topLevelTest (void *args)
{
	if(multinode_test_main(args)<0) {
      printf("\n multinode_test_main failed!!");
	}

    pthread_exit((void*) 0);
    return NULL;
}

